#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../siod/siod.h"
#include "../common/common.h"
#include "../common/cmalloc.h"
#include "types.h"
#include "calc.h"

/* Saving strategy:
   We always have the fallback option to save as a floating-point
   or string constant. But it is desirable to save formulas as such
   to not lose information. In particular, a wk1 file that can be
   loaded by SIAG should also be possible to save again.
   The formula string is parsed by SIOD into list form using the
   read-from-string function. The resulting list is then passed
   into a C function, which traverses it and generates RPN code
   to be stored in the worksheet.
*/

static LISP pre2post(LISP l)
{
	LISP p;
	if (CONSP(l)) {
		/* process all the arguments first; must also count args */
		for (p = CDR(l); NNULLP(p); p = CDR(p)) {
			pre2post(CAR(p));
		}
		if (SYMBOLP(CAR(l))) {
			/* should save the matching operator code */
			printf("operator: %s\n", get_c_string(CAR(l)));
		} else {
			printf("error: missing operator\n");
		}
	} else if (FLONUMP(l)) {
		/* simple enough */
		printf("number: %g\n", get_c_double(l));
	} else if (SYMBOLP(l)) {
		/* this will be a range, a reference or an error */
		printf("symbol: %s\n", get_c_string(l));
	} else {
		printf("error: missing operand\n");
	}
	return NIL;
}

static int col_last_changed(buffer *b, int row)
{
        int i;

        if (row > b->alloc_lines) return 0;
        for (i = b->alloc_cols[row]; i > 1; i--)
                if (ret_type(b, row, i) != EMPTY)
                        break;
        return i;
}

static void writeint(unsigned char *p, int n)
{
	p[0] = n % 256;
	p[1] = n / 256;
}

static void writefloat(unsigned char *p, double x)
{
	memcpy(p, &x, 8);	/* FIXME */
}

static int save_cell(FILE *fp, buffer *buf, long r, long c)
{
	unsigned char b[2000], s[2000], *p;
	double num;
	int len;

	switch (ret_type(buf, r, c)) {
	case LABEL:
		writeint(b, 0x0F);		/* type LABEL */
		p = (unsigned char *)ret_text(buf, r, c);
		len = 1+2+2+1+strlen((char *)p)+1;
		writeint(b+2, len);
		b[4] = 0xFF;			/* format */
		writeint(b+5, c-1);		/* column */
		writeint(b+7, r-1);		/* row */
		b[9] = '\'';
		strcpy((char *)b+10, (char *)p);
		fwrite(b, 1, len+4, fp);
		break;
	case STRING:	/* do later */
		break;
	case EXPRESSION:
		writeint(b, 0x0E);		/* type NUMBER */
		len = 1+2+2+8;
		writeint(b+2, len);		/* length 1+2+2+8 */
		b[4] = 0xFF;			/* format */
		writeint(b+5, c-1);		/* column */
		writeint(b+7, r-1);		/* row */
		num = ret_val(buf, r, c).number;
		writefloat(b+9, num);
		sprintf((char *)s, "(scm->list %ld %ld)", r, c);
		execute((char *)s);
		fwrite(b, 1, len+4, fp);	/* 4 is for type and length */
		break;
	default:	/* ignore for now */
		break;
	}
	return 0;
}

static int save_wk1(char *fn, buffer *buf)
{
	long r, c;
	FILE *fp;
	static unsigned char preblurb[] =
		{0x00, 0x00, 0x02, 0x00, 0x06, 0x04};
	static unsigned char postblurb[] =
		{0x01, 0x00, 0x00, 0x00};

	if ((fp = fopen(fn, "wb")) == NULL)
		return 1;

	if (fwrite(preblurb, 1, sizeof preblurb, fp) != sizeof preblurb) {
		fclose(fp);
		return 1;
	}
	for (r = 1; r <= line_last_used(buf); r++) {
		for (c = 1; c <= col_last_changed(buf, r); c++) {
			if (save_cell(fp, buf, r, c)) {
				fclose(fp);
				return 1;
			}
		}
	}
	if (fwrite(postblurb, 1, sizeof postblurb, fp) != sizeof postblurb) {
		fclose(fp);
		return 1;
	}
	fclose(fp);
	return 0;
}

/* Loading strategy:
   The Lotus format stores formulas in postfix notation. It is most
   straightforward to translate this to prefix, i.e. Scheme.
   Every token is stored on a stack, in string form. Example:
   1 2 + => "1", "1" "2", "(+ 1 2)"
   So an operator is done by creating a new string from several on
   the stack.
*/

static char *stack[10];
static int sp;

static void push(char *p)
{
TRACEME((f, "push(%s)", p));
	if (sp >= 10) return;
	stack[sp++] = p;
}

static char *pop(void)
{
TRACEME((f, "pop() => %s", stack[sp-1]));
	if (sp <= 0) return NULL;
	--sp;
	return stack[sp];
}

/* borrowed from teapot */
static double readfloat(const unsigned char *s)
{
  double x;
  int sg,e,i;

  x=0.0;
  for (i=1; i<256; i<<=1) x=x/2.0+!!(s[0]&i);
  for (i=1; i<256; i<<=1) x=x/2.0+!!(s[1]&i);
  for (i=1; i<256; i<<=1) x=x/2.0+!!(s[2]&i);
  for (i=1; i<256; i<<=1) x=x/2.0+!!(s[3]&i);
  for (i=1; i<256; i<<=1) x=x/2.0+!!(s[4]&i);
  for (i=1; i<256; i<<=1) x=x/2.0+!!(s[5]&i);
  x=x/2.0+!!(s[6]&0x01);
  x=x/2.0+!!(s[6]&0x02);
  x=x/2.0+!!(s[6]&0x04);
  x=x/2.0+!!(s[6]&0x08);
  x=x/2.0+1.0;
  if ((e=((s[6]>>4)+((s[7]&0x7f)<<4))-1023)==-1023)
  {
    x=0.0;
    e=0;
  }
  if (s[7]&0x80) sg=-1; else sg=1;
  return (sg*ldexp(x,e));
}

static int readint(unsigned char *p)
{
	return p[0]+256*p[1];
}

static int makecoord(int orig, int coord)
{
        if (coord & 0x8000)
                /* relative; this requires a 2-complement int */
                coord += orig;
        return coord & 0x3FFF;
}

#define CASE0(n,s,m) \
	case (n): \
		sprintf(b, "(%s)", s); \
		push(cstrdup(b)); \
		i += (m); \
		break;

#define CASE1(n,s,m) \
	case (n): \
		p1 = pop(); \
		sprintf(b, "(%s %s)", s, p1); cfree(p1); \
		push(cstrdup(b)); \
		i += (m); \
		break;

#define CASE2(n,s,m) \
	case (n): \
		p1 = pop(); q1 = pop(); \
		sprintf(b, "(%s %s %s)", s, q1, p1); cfree(p1); cfree(q1); \
		push(cstrdup(b)); \
		i += (m); \
		break;

#define CASEn(n,s,m) \
        case (n): \
                sprintf(b, "(%s", s); \
                addargs(b, p[i+1]); \
                strcat(b, ")"); \
                push(cstrdup(b)); \
                i += (m); \
                break;

static void addargs(char *b, int n)
{
        if (n) {
                char *p = pop();
                addargs(b, n-1);
                strcat(b, " ");
                strcat(b, p);
        }
}

static void formula(int len, unsigned char *p, int r, int c)
{
	int i = 0;
	char b[2000];
	char *p1, *q1;

	sp = 0;

	while (i < len) {
		switch (p[i]) {
		case 0x00:
			sprintf(b, "%f", readfloat(p+i+1));
			push(cstrdup(b));
			i += 9;
			break;
		case 0x01:
			sprintf(b, "r%dc%d",
                                1+makecoord(r, readint(p+i+3)),
                                1+makecoord(c, readint(p+i+1)));
			push(cstrdup(b));
			i += 5;
			break;
		case 0x02:
			sprintf(b, "r%dc%d..r%dc%d",
                                1+makecoord(r, readint(p+i+3)),
                                1+makecoord(c, readint(p+i+1)),
                                1+makecoord(r, readint(p+i+7)),
                                1+makecoord(c, readint(p+i+5)));
			push(cstrdup(b));
			i += 9;
			break;
		case 0x03:
			return;
		case 0x04:
			i += 1;
			break;
		case 0x05:
			sprintf(b, "%d", readint(p+i+1));
			push(cstrdup(b));
			i += 3;
			break;
		case 0x06:
			i++;
			p1 = b;
			*p1++ = '"';
			while (p[i]) {
				if (p[i] == '"' || p[i] == '\\') *p1++ = '\\';
				*p1++ = p[i++];
			}
			*p1++ = '"';
			*p1 = '\0';
			push(cstrdup(b));
			i++;
			break;
		CASE1(0x08, "-", 1);
		CASE2(0x09, "+", 1);
		CASE2(0x0A, "-", 1);
		CASE2(0x0B, "*", 1);
		CASE2(0x0C, "/", 1);
		CASE2(0x0D, "exp", 1);
		CASE2(0x0E, "=", 1);
		CASE2(0x0F, "!=", 1);
		CASE2(0x10, "<=", 1);
		CASE2(0x11, ">=", 1);
		CASE2(0x12, "<", 1);
		CASE2(0x13, ">", 1);
		CASE2(0x14, "bit-and", 1);
		CASE2(0x15, "bit-or", 1);
		CASE1(0x16, "bit-not", 1);
		CASE1(0x17, "+", 1);
		CASE1(0x1F, "@NA", 1);
		CASE1(0x20, "@ERR", 1);
		CASE1(0x21, "@ABS", 1);
		CASE1(0x22, "@INT", 1);
		CASE1(0x23, "@SQRT", 1);
		CASE1(0x24, "@LOG", 1);
		CASE1(0x25, "@LN", 1);
		CASE0(0x26, "@PI", 1);
		CASE1(0x27, "@SIN", 1);
		CASE1(0x28, "@COS", 1);
		CASE1(0x29, "@TAN", 1);
		CASE1(0x2A, "@ATAN2", 1);
		CASE1(0x2B, "@ATAN", 1);
		CASE1(0x2C, "@ASIN", 1);
		CASE1(0x2D, "@ACOS", 1);
		CASE1(0x2E, "@EXP", 1);
		CASE2(0x2F, "@MOD", 1);
		CASE0(0x30, "@CHOOSE", 1);
		CASE0(0x31, "@ISNA", 1);
		CASE0(0x32, "@ISERR", 1);
		CASE0(0x33, "@FALSE", 1);
		CASE0(0x34, "@TRUE", 1);
		CASE0(0x35, "@RAND", 1);
		CASE0(0x36, "@DATE", 1);
		CASE0(0x37, "@TODAY", 1);
		CASE0(0x38, "@PMT", 1);
		CASE0(0x39, "@PV", 1);
		CASE0(0x3A, "@FV", 1);
		CASE0(0x3B, "@IF", 1);
		CASE0(0x3C, "@DAY", 1);
		CASE0(0x3D, "@MONTH", 1);
		CASE0(0x3E, "@YEAR", 1);
		CASE1(0x3F, "@ROUND", 1);
		CASE0(0x40, "@TIME", 1);
		CASE0(0x41, "@HOUR", 1);
		CASE0(0x42, "@MINUTE", 1);
		CASE0(0x43, "@SECOND", 1);
		CASE1(0x44, "@ISNUMBER", 1);
		CASE1(0x45, "@ISSTRING", 1);
		CASE1(0x46, "@LENGTH", 1);
		CASE1(0x47, "@VALUE", 1);
		CASE1(0x48, "@FIXED", 1);
		CASE1(0x49, "@MID", 1);
		CASE1(0x4A, "@CHR", 1);
		CASE1(0x4B, "@ASCII", 1);
		CASE1(0x4C, "@FIND", 1);
		CASE0(0x4D, "@DATEVALUE", 1);
		CASE0(0x4E, "@TIMEVALUE", 1);
		CASE0(0x4F, "@CELLPOINTER", 1);
        /* 50-54 take an additional byte for the number of args */
                CASEn(0x50, "@SUM", 2);
                CASEn(0x51, "@AVG", 2);
                CASEn(0x52, "@CNT", 2);
                CASEn(0x53, "@MIN", 2);
                CASEn(0x54, "@MAX", 2);
		CASE1(0x55, "@VLOOKUP", 1);
		CASE1(0x56, "@NPV", 1);
		CASE1(0x57, "@VAR", 1);
		CASE1(0x58, "@STD", 1);
		CASE1(0x59, "@IRR", 1);
		CASE1(0x5A, "@HLOOKUP", 1);
		CASE1(0x5B, "DSUM", 1);
		CASE1(0x5C, "AVG", 1);
		CASE1(0x5D, "DCNT", 1);
		CASE1(0x5E, "DMIN", 1);
		CASE1(0x5F, "DMAX", 1);
		CASE1(0x60, "DVAR", 1);
		CASE1(0x61, "DSTD", 1);
		CASE1(0x62, "@INDEX", 1);
		CASE1(0x63, "@COLS", 1);
		CASE1(0x64, "@ROWS", 1);
		CASE1(0x65, "@REPEAT", 1);
		CASE1(0x66, "@UPPER", 1);
		CASE1(0x67, "@LOWER", 1);
		CASE0(0x68, "@LEFT", 1);
		CASE0(0x69, "@RIGHT", 1);
		CASE1(0x6A, "@REPLACE", 1);
		CASE1(0x6B, "@PROPER", 1);
		CASE1(0x6C, "@CELL", 1);
		CASE1(0x6D, "@TRIM", 1);
		CASE1(0x6E, "@CLEAN", 1);
		CASE2(0x71, "@STREQ", 1);
		CASE1(0x72, "@CALL", 1);
		CASE1(0x74, "@RATE", 1);
		CASE1(0x75, "@TERM", 1);
		CASE1(0x76, "@CTERM", 1);
		CASE1(0x77, "@SLN", 1);
		CASE1(0x78, "@SOY", 1);
		CASE1(0x79, "@DDB", 1);
		CASE1(0x9C, "@AAFSTART", 1);
		CASE1(0xCE, "@AAFUNKNOWN", 1);
		CASE1(0xFF, "@AAFEND", 1);
		default:
			return;
		}
	}
}

static void wk1(FILE *fp, buffer *buf)
{
	char b[2000];
	char *p1;
	cval value;

	for (;;) {
		unsigned char head[4];
		unsigned char *p;
		int type, len;
		if (fread(head, 1, 4, fp) != 4) {
			fprintf(stderr, "Unexpected end of file\n");
			return; /* error or end */
		}
		type = readint(head);
		len = readint(head+2);
		p = malloc(len);
		if (!p) {
			fprintf(stderr, "Couldn't allocate %d bytes\n", len);
			return;
		}
		if (fread(p, 1, len, fp) != len) {
			fprintf(stderr, "Couldn't read %d bytes\n", len);
			free(p);
			return;
		}
		switch (type) {
		case 0x00:
			break;
		case 0x01:
			free(p);
			return;
		case 0x08:
			break;
		case 0x0D:
			value.number = readint(p+5);
			sprintf(b, "%d", readint(p+5));
			ins_data(buf, siod_interpreter, b,
				value, EXPRESSION,
				1+readint(p+3), 1+readint(p+1));
			break;
		case 0x0E:
			value.number = readfloat(p+5);
			sprintf(b, "%f", value.number);
			ins_data(buf, siod_interpreter, b,
				value, EXPRESSION,
				1+readint(p+3), 1+readint(p+1));
			break;
		case 0x0F:
			switch (p[5]) {
			case '\'':
				break;
			case '"':
				break;
			case '^':
				break;
			default:
				break;
			}
			value.number = 0;
			ins_data(buf, siod_interpreter, (char *)p+6,
				value, LABEL,
				1+readint(p+3), 1+readint(p+1));
			break;
		case 0x10:
		/* 5-12 = value */
		/* 13-14 = formula length */
			formula(readint(p+13), p+15,
				readint(p+3), readint(p+1));
			p1 = pop();
			value.number = readfloat(p+5);
			ins_data(buf, siod_interpreter, p1,
				value, EXPRESSION,
				1+readint(p+3), 1+readint(p+1));
			cfree(p1);
			break;
		default:
			break;
		}
		free(p);
	}
}

static int load_wk1(char *fn, buffer *buf)
{
	FILE *fp;
	if ((fp = fopen(fn, "rb")) == NULL)
		return 1;

	wk1(fp, buf);
	fclose(fp);
	return 0;
}

/* File format guessing:
   1. extension .wk1 or .wks
   2. starts with the sequence 0x00 0x00 0x02 0x00
*/
static int myformat(char *fn)
{
        char *ext;
        FILE *fp;
        char b[10];

        ext = strrchr(fn, '.');
        if (!ext) return 0;     /* no extension */
        if (cstrcasecmp(ext, ".wk1") && cstrcasecmp(ext, ".wks"))
                return 0;       /* wrong extension */
        if ((fp = fopen(fn, "rb")) == NULL) return 0;    /* can't open */
        if (fread(b, 1, 4, fp) != 4) {		/* can't read */
		fclose(fp);
		return 0;
	}
	if (b[0] == 0 && b[1] == 0 && b[2] == 2 && b[3] == 0) {	/* gotcha */
		fclose(fp);
		return 1;
	}
        fclose(fp);
        return 0;       /* yep, sure is ;-) */
}

void fileio_wk1_init(void)
{
	init_subr_1("pre->post", pre2post);
        register_format(load_wk1, save_wk1, myformat, "Lotus 1-2-3 (*.wk1)");
}

